#include <avr/io.h>
#include "lcd-routines.h"
#include "wait1000ms.h"
#include "config.h"
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <string.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <math.h>
//#define DebugOut 3		// wenn gesetzt, Zwischenausgaben bei Widerstandmessung in Zeile 2
//#define DebugOut 4		// wenn gesetzt, Spannungsausgabe bei Diodenmessung in Zeilen 3+4
//#define DebugOut 5		// wenn gesetzt, Zwischenausgaben der TransistorUntersuchung in Zeile 2+3
//#define DebugOut 10		// wenn gesetzt, Zwischenausgaben in ReadCapacity in Zeilen 3+4 
/*########################################################################################
	Konfigurations-Einstellungen
*/

/* Port fr die Test-Pins
  Dieser Port muss ber einen ADC verfgen (beim Mega8 also PORTC).
  Fr die Test-Pins mssen die unteren 3 Pins dieses Ports benutzt werden.
  Bitte die Definitionen fr TP1, TP2 und TP3 nicht ndern!
*/

#define ADC_PORT PORTC
#define ADC_DDR DDRC
#define ADC_PIN PINC
#define TP1 PC0
#define TP2 PC1
#define TP3 PC2

  /*Einstellungen fr Kapazittsmessung (nur fr ATMega8 interessant)
  Der Test, ob ein Kondensator vorhanden ist, dauert relativ lange, mit ber 50ms je Testvorgang ist zu rechnen
  Bei allen 6 mglichen Testvorgngen ergibt das eine Verlngerung der Testdauer um ca 0,3s bis 0,5s.
  Mit CAP_TEST_MODE lassen sich die durchgefhrten Tests festlegen.

  Bedeutungen der Bits (7 = MSB):
  7:6 Nicht verwendet

  5:4 Test-Modus
  00: Kondensator-Messung deaktiviert
  01: Kondensator-Messung fr eine einstellbare Pin-Kombination (in beide Richtungen); verlngert Testdauer um ca. 120...200ms
  10: Kondensator-Messung fr alle 6 Pin-Kombinationen; verlngert Testdauer um ca. 300...500ms
  
  3:2 Erster Pin der gewhlten Pin-Kombination (0...2), nur entscheidend wenn Bits 5:4 = 01

  1:0 Zweiter Pin der gewhlten Pin-Kombination (0...2), nur entscheidend wenn Bits 5:4 = 01
  */
  #define CapTestMode  0x22         //Messung fr alle 6 Pin-Kombinationen



/*
  Genaue Werte der verwendeten Widerstnde in Ohm.
  Der Nennwert fr R_L ist 680 Ohm, fr R_H 470kOhm.
  
  Um das Programm auf Abweichungen von diesen Werten (z.B. durch Bauteiltoleranzen)
  zu kalibrieren, die Widerstandswerte in Ohm in die folgenden Defines eintragen:
*/
  #define R_L_VAL 680		// Normwert 680 Ohm
  #define R_H_VAL 4700		// Normwert 470000 Ohm, durch 100 dividiert

  #define PIN_RP  22		// Geschtzter Innenwiderstand PORT Ausgang VCC
  #define PIN_RM  18		// geschtzter Innenwiderstand PORT Ausgang GND


  /*	Faktoren fr die Kapazittsmessung bei Kondensatoren
    Diese Faktoren hngen von Fertigungstoleranzen des AVR ab und mssen somit ggf. angepasst werden
    H_CAPACITY_FACTOR ist fr die Messung mit 470k-Widerstand (geringe Kapazitt)
    L_CAPACITY_FACTOR ist fr die Messung mit 680-Ohm-Widerstand (hohe Kapazitt)
  Gemessen wird die Zeit (t) in s bis zum Erreichen der Vergleichsspannung (1.3V)
  des analogen Komparators.
  Die Spannung am Kondensator steigt nach Uc = VCC * (1 - e**(-t/T))
  Die 1.3V werden somit nach t = -ln(3.7/5)*T  = 0.3011*T erreicht. 
  Zeitkonstante T = R * C ; also
  C = T / R
  fr die 470 kOhm  ergibt sich nach C = t / (0.3011 * 470000)
  H_Fakt = 707/100 fr das Ergebnis in pF

  Fr die 680 Ohm sind die Verhltnisse komplizierter.
  zum Zeitpunkt des Komparator-Schaltens fliesst noch deutlich Strom, der
  einen zustzlichen Spannungsanstieg des LowPin zur Folge hat.
   = (VCC - 1.3V) / (R_L + PIN_RP) * PIN_RM = 105mV
  Der Komparator schaltet also bei einer Kondensator-Spannung von etwa 1.2V!!!
  Das ergibt als Schaltzeit t = -ln((5-1.2)/5)*T = 0.274437*T .
  fr die 680 Ohm ergibt sich nach C = t / (0.274437 * (680 + PIN_RM + PIN_RP))
  oder C = Z * 5.060869   (253/50)
  
    Der gesamte Messbereich ist ca. 100pF bis 80mF.
  */
//#define H_CAPACITY_FACTOR  707
//#define H_CAPACITY_FACTOR  724    //ersetzt durch RHmultip
//#define L_CAPACITY_FACTOR  253
//#define L_CAPACITY_FACTOR  273	//ersetzt durch RLmultip
#ifdef USE_EEPROM
#define MEM_TEXT EEMEM
#else
#define MEM_TEXT PROGMEM
#endif

// Widerstand 680 Ohm         200  250  300  350  400  450  500  550  600  650  700  750  800  850  900  950 1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
uint16_t RLtab[] MEM_TEXT = {3402,2708,2245,1914,1666,1473,1318,1192,1086, 997, 921, 855, 797, 745, 700, 659, 622, 589, 559, 531, 506, 483, 461, 441, 423};

#ifdef WITH_AUTO_REF

// Widerstand 680 Ohm        1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
//uint16_t RLtab[] MEM_TEXT = { 702, 660, 622, 588, 558, 529, 504, 480, 458};
//Widerstand 470000 Ohm      1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
uint16_t RHtab[] MEM_TEXT = { 954, 903, 856, 814, 775, 740, 707, 676, 648};

#endif

// durch Faktoren wird der ADC-Wert so umgerechnet, dass die Spannungen in mV gemessen werden!
// alle Abfragen wurden durch solche mV Abfragen ersetzt.
#define U_VCC 5001		// Maximalwert der ADC-Messung


/*########################################################################################
Ende der Konfigurations-Einstellungen
*/

/*Strings im PROGMEM
Beim Hinzufgen weiter Sprachen bitte auf die Lnge achten ,
sonst gibt es Probleme mit der LCD-Anzeige bei den verschiedenen Sprachen!
*/

#if defined(LANG_GERMAN)		//deutsch
  unsigned char TestRunning[] MEM_TEXT = "Testen...";
  unsigned char BatWeak[] MEM_TEXT = "schwach";
  unsigned char BatEmpty[] MEM_TEXT = "leer!";
  unsigned char TestFailed2[] MEM_TEXT = "defektes ";
  unsigned char Bauteil[] MEM_TEXT = "Bauteil";
  unsigned char Diode[] MEM_TEXT = "Diode: ";
  unsigned char GAK[] MEM_TEXT = "GAK=";
  unsigned char Triac[] MEM_TEXT = "Triac";
  unsigned char Thyristor[] MEM_TEXT = "Thyristor";
  unsigned char Unknown[] MEM_TEXT = " unbek.";
  unsigned char TestFailed1[] MEM_TEXT = "Kein,unbek. oder";
  unsigned char OrBroken[] MEM_TEXT = "oder defekt ";
#endif

#if defined(LANG_ENGLISH)		//english
  unsigned char TestRunning[] MEM_TEXT = "Testing...";
  unsigned char BatWeak[] MEM_TEXT = "weak";
  unsigned char BatEmpty[] MEM_TEXT = "empty!";
  unsigned char TestFailed2[] MEM_TEXT = "damaged ";
  unsigned char Bauteil[] MEM_TEXT = "part";
  unsigned char Diode[] MEM_TEXT = "Diode: ";
  unsigned char GAK[] MEM_TEXT = "GAC=";
  unsigned char Triac[] MEM_TEXT = "Triac";
  unsigned char Thyristor[] MEM_TEXT = "Thyristor";
  unsigned char Unknown[] MEM_TEXT = " unknown";
  unsigned char TestFailed1[] MEM_TEXT = "No, unknown, or";
  unsigned char OrBroken[] MEM_TEXT = "or damaged ";
#endif

#if defined(LANG_POLISH)		//polnisch
  unsigned char TestRunning[] MEM_TEXT = "Testowanie...";
  unsigned char BatWeak[] MEM_TEXT = "slaba";
  unsigned char BatEmpty[] MEM_TEXT = "za slaba";
  unsigned char TestFailed2[] MEM_TEXT = "lub uszkodz.";
  unsigned char Bauteil[] MEM_TEXT = "Element";
  unsigned char Diode[] MEM_TEXT = "Dioda: ";
  unsigned char GAK[] MEM_TEXT = "GAK=";
  unsigned char Triac[] MEM_TEXT = "Triak";
  unsigned char Thyristor[] MEM_TEXT = "Tyrystor";
  unsigned char Unknown[] MEM_TEXT = " nieznany";
  unsigned char TestFailed1[] MEM_TEXT = "brak elementu";
  unsigned char OrBroken[] MEM_TEXT = "lub uszkodz. ";
#endif

#if defined(LANG_CZECH)		//Tschechisch
  unsigned char TestRunning[] MEM_TEXT = "Probiha mereni...";
  unsigned char BatWeak[] MEM_TEXT = "slaba";
  unsigned char BatEmpty[] MEM_TEXT = "prazdna!";
  unsigned char TestFailed2[] MEM_TEXT = "vadna ";
  unsigned char Bauteil[] MEM_TEXT = "soucastka";
  unsigned char Diode[] MEM_TEXT = "Dioda: ";
  unsigned char GAK[] MEM_TEXT = "GAK=";
  unsigned char Triac[] MEM_TEXT = "Triak";
  unsigned char Thyristor[] MEM_TEXT = "Tyristor";
  unsigned char Unknown[] MEM_TEXT = " neznama";
  unsigned char TestFailed1[] MEM_TEXT = "Zadna, neznama";
  unsigned char OrBroken[] MEM_TEXT = "nebo vadna ";
#endif

#if defined(LANG_SLOVAK)		//slowakisch
  unsigned char TestRunning[] MEM_TEXT = "PREBIEHA TEST";
  unsigned char BatWeak[] MEM_TEXT = "slaba";
  unsigned char BatEmpty[] MEM_TEXT = "prazdna!";
  unsigned char TestFailed2[] MEM_TEXT = "vadna ";
  unsigned char Bauteil[] MEM_TEXT = "suciastka!";
  unsigned char Diode[] MEM_TEXT = "Dioda: ";
  unsigned char GAK[] MEM_TEXT = "GAK=";
  unsigned char Triac[] MEM_TEXT = "Triak";
  unsigned char Thyristor[] MEM_TEXT = "Tyristor";
  unsigned char Unknown[] MEM_TEXT = " neznama";
  unsigned char TestFailed1[] MEM_TEXT = "Ziadna, neznama";
  unsigned char OrBroken[] MEM_TEXT = "alebo vadna ";
#endif



//Sprachunabhngige Strings
unsigned char Bat[] MEM_TEXT = "Bat. ";
unsigned char OK[] MEM_TEXT = "OK";
unsigned char mosfet[] MEM_TEXT = "-MOS";
unsigned char jfet[] MEM_TEXT = "JFET";
unsigned char GateCap[] MEM_TEXT = " C=";
unsigned char hfestr[] MEM_TEXT ="B=";
unsigned char NPN[] MEM_TEXT = "NPN ";
unsigned char PNP[] MEM_TEXT = "PNP ";
unsigned char ebcstr[] MEM_TEXT = " EBC=";
unsigned char gds[] MEM_TEXT = "GDS=";
unsigned char Uf[] MEM_TEXT = "Uf=";
unsigned char vt[] MEM_TEXT = "Vt=";
#ifdef C_MESS
unsigned char CapZeich[] MEM_TEXT = {'-',LCD_CHAR_CAP,'-',0};
#endif
unsigned char AnKat[] MEM_TEXT = {'-', LCD_CHAR_DIODE1, '-',0};
unsigned char KatAn[] MEM_TEXT = {'-', LCD_CHAR_DIODE2, '-',0};
unsigned char Dioden[] MEM_TEXT = {'*',LCD_CHAR_DIODE1, ' ', ' ',0};
#ifdef R_MESS
unsigned char Resis[] MEM_TEXT = {'-', LCD_CHAR_RESIS1, LCD_CHAR_RESIS2,'-',0};
#endif
unsigned char TestTimedOut[] MEM_TEXT = "Timeout!";

#ifdef WITH_SELFTEST
unsigned char mVT[] MEM_TEXT = "mV ";
unsigned char URefT[] MEM_TEXT = "Ref=";
unsigned char RHfakt[] MEM_TEXT = "RHf=";
unsigned char RLfakt[] MEM_TEXT = " RLf=";
unsigned char ATE[] MEM_TEXT = "Auto Test End";
unsigned char VERSION[] MEM_TEXT = "Version 0.92k";
unsigned char SELFTEST[] MEM_TEXT = "Selftest mode..";
unsigned char RH1L[] MEM_TEXT = "RH-";
unsigned char RH1H[] MEM_TEXT = "RH+";
unsigned char RILO[] MEM_TEXT = "Ri_Lo= (mV)";
unsigned char RIHI[] MEM_TEXT = "Ri_Hi= (mV)";
unsigned char RLRL[] MEM_TEXT = "L1+2- .3- 2+.";
unsigned char RHRH[] MEM_TEXT =     "H1+2- .3- 2+.";
unsigned char RELPROBE[] MEM_TEXT = "isolate probe";
 #define MULTIP
 #define LCD_CLEAR
#endif
#ifdef DebugOut
 #define LCD_CLEAR
#endif

unsigned char DiodeIcon1[] MEM_TEXT = { 0x11,
				     0x19,
				     0x1d,
				     0x1f,
				     0x1d,
				     0x19,
				     0x11,
				     0x00};	// Diode-Icon Anode links

unsigned char DiodeIcon2[] MEM_TEXT = { 0x11,
				     0x13,
				     0x17,
				     0x1f,
				     0x17,
				     0x13,
				     0x11,
				     0x00};	// Diode-Icon Anode rechts
#ifdef C_MESS
unsigned char CapIcon[] MEM_TEXT = { 0x1b,
				     0x1b,
				     0x1b,
				     0x1b,
				     0x1b,
				     0x1b,
				     0x1b,
				     0x00};	// Capacitor Icon
#endif
#ifdef R_MESS
unsigned char ResIcon1[] MEM_TEXT = { 0x00,
				     0x0f,
				     0x08,
				     0x18,
				     0x08,
				     0x0f,
				     0x00,
				     0x00};	// Resistor Icon1
unsigned char ResIcon2[] MEM_TEXT = { 0x00,
				     0x1e,
				     0x02,
				     0x03,
				     0x02,
				     0x1e,
				     0x00,
				     0x00};	// Resistor Icon2
#endif

unsigned char PinRLtab[] MEM_TEXT = { (1<<(TP1*2)),
				     (1<<(TP2*2)),
				     (1<<(TP3*2))};	// Tabelle der R-L Schaltbefehle Pin 0,1,2

unsigned char PinADCtab[] MEM_TEXT = { (1<<TP1),
				     (1<<TP2),
				     (1<<TP3)};	// Tabelle der ADC-Pins Schaltbefehle Pin 0,1,2

#ifdef LCD_CYRILLIC	//Omega- und -Zeichen als Custom-Zeichen erzeugen, weil diese Zeichen im kyrillischen Zeichensatz nicht enthalten sind
  unsigned char CyrillicOmegaIcon[] MEM_TEXT = {0,0,14,17,17,10,27,0};	//Omega
  unsigned char CyrillicMuIcon[] MEM_TEXT = {0,17,17,17,19,29,16,16};	//
#endif

//Ende der EEPROM-Strings

//Watchdog
#define WDT_enabled
/* Wird das Define "WDT_enabled" entfernt, wird der Watchdog beim Programmstart
 nicht mehr aktiviert. Das ist fr Test- und Debuggingzwecke sinnvoll.
 Fr den normalen Einsatz des Testers sollte der Watchdog aber unbedingt aktiviert werden!
*/


struct Diode {
  uint8_t Anode;
  uint8_t Cathode;
  unsigned int Voltage;
};

void CheckPins(uint8_t HighPin, uint8_t LowPin, uint8_t TristatePin);
void ChargePin10ms(uint8_t PinToCharge, uint8_t ChargeDirection);
unsigned int ReadADC(uint8_t mux);		// LeseRoutine fr ADC
unsigned int W5msReadADC(uint8_t mux);		// wartet 5ms und liest dann ADC
unsigned int W20msReadADC(uint8_t mux);		// wait 5ms and read ADC
void lcd_show_format_cap(void);
void ReadCapacity(uint8_t HighPin, uint8_t LowPin);		//Kapazittsmessung nur auf Mega8 verfgbar
void UfAusgabe(uint8_t bcdchar);	// Ausgabe der Schwellwertspannung(en) Uf
void RvalOut(uint8_t ii);		// Ausgabe der Widerstandswerte
void EntladePins();			// Kondensatoren entladen
void RefVoltage();			// Kompensiere Referenz-Spannung fr Komparator
void AutoCheck();			// Selbst-Prfung
unsigned int getRLmultip(unsigned int cvolt);  // get C-Multiplikator fr Spannung cvolt

#define R_DDR DDRB
#define R_PORT PORTB

/* Port fr die Testwiderstnde
  Die Widerstnde mssen an die unteren 6 Pins des Ports angeschlossen werden,
  und zwar in folgender Reihenfolge:
  RLx = 680R-Widerstand fr Test-Pin x
  RHx = 470k-Widerstand fr Test-Pin x

  RL1 an Pin 0
  RH1 an Pin 1
  RL2 an Pin 2
  RH2 an Pin 3
  RL3 an Pin 4
  RH3 an Pin 5

*/



#define ON_DDR DDRD
#define ON_PORT PORTD
#define ON_PIN_REG PIND
#define ON_PIN PD6	//Pin, der auf high gezogen werden muss, um Schaltung in Betrieb zu halten
#define RST_PIN PD7	//Pin, der auf low gezogen wird, wenn der Einschalt-Taster gedrckt wird

//Bauteile
#define PART_NONE 0
#define PART_DIODE 1
#define PART_TRANSISTOR 2
#define PART_FET 3
#define PART_TRIAC 4
#define PART_THYRISTOR 5
#define PART_RESISTOR 6
#define PART_CAPACITOR 7

//Ende (Bauteile)
//Spezielle Definitionen fr Bauteile
//FETs
#define PART_MODE_N_E_MOS 1
#define PART_MODE_P_E_MOS 2
#define PART_MODE_N_D_MOS 3
#define PART_MODE_P_D_MOS 4
#define PART_MODE_N_JFET 5
#define PART_MODE_P_JFET 6

//Bipolar
#define PART_MODE_NPN 1
#define PART_MODE_PNP 2

// die folgenden Anweisungen sind fr die korrekte Einstellung des ADC-Clock Vorteilers
// Passend fr 1MHz, 2MHz, 4MHz, 8MHz und 16MHz
#if F_CPU/125000 == 8
 #define AUTO_CLOCK_DIV (1<<ADPS1) | (1<<ADPS0)
#endif
#if F_CPU/125000 == 16
 #define AUTO_CLOCK_DIV (1<<ADPS2)
#endif
#if F_CPU/125000 == 32
 #define AUTO_CLOCK_DIV (1<<ADPS2) | (1<<ADPS0)
#endif
#if F_CPU/125000 == 64
 #define AUTO_CLOCK_DIV (1<<ADPS2) | (1<<ADPS1)
#endif
#if F_CPU/125000 == 128
 #define AUTO_CLOCK_DIV (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0)
#endif


struct Diode diodes[6];
uint8_t NumOfDiodes;

struct {
  unsigned int hfe[2];		//Verstrkungsfaktor
  //unsigned int uBE[2];		//B-E-Spannung fr Transistor
  uint8_t b,c,e;			//Anschlsse des Transistors
}trans;
unsigned int gthvoltage;	//Gate-Schwellspannung

uint8_t PartReady;		//Bauteil fertig erkannt
uint8_t PartMode;
uint8_t tmpval, tmpval2;

#ifdef WITH_UART
 #define TXD_MSK (1<<TxD)
#else
 #define TXD_MSK 0
#endif

#ifdef SWUART_INVERT
  #define TXD_VAL 0
#else
  #define TXD_VAL TXD_MSK
#endif

#ifdef R_MESS
struct resis_t{
   unsigned long rx;		// Widerstandswert RX  
   uint8_t kfl;			// 1, wenn Messung mit R_L, sonst 100
   uint8_t ra,rb;			// Pinne von RX
   uint8_t rt;			// Tristate-Pin
} resis[3];
 unsigned int rxv, rtst;
 unsigned int rvmax;
 uint8_t NumOfR;		//Anzahl der gefundenen Widerstnde
#endif


#ifdef C_MESS 
  unsigned char C_Prefix_tab[] PROGMEM = { 'p','n',LCD_CHAR_U,'m'}; // pF,nF,F,mF
  #define MULTIP
#endif
#ifdef MULTIP
// Multiplikator fr die Kapazittsmessung mit R_H (470KOhm)
  unsigned int RHmultip=DEFAULT_RH_C_FACT;
// Multiplikator fr die Kapazittsmessung mit R_L (680 Ohm)
  unsigned int RLmultip=DEFAULT_RL_C_FACT;
#endif


uint8_t ii;			// Hilfszhler
unsigned long cval;		// gemessener Kapazittswert
uint8_t cpre;			//Prefix fr Farad Angabe 0=p, 1=n, 2=, 3=m
uint8_t ca, cb;			//Kondensator-Pins

uint8_t PartFound;	 	//das gefundene Bauteil
char outval[12];		// String fr ASCII-Ausgabe


//Programmbeginn
int main(void) {
  //Einschalten
  ON_DDR = (1<<ON_PIN);
  ON_PORT = (1<<ON_PIN) | (1<<RST_PIN);	//Strom an und Pullup fr Reset-Pin
  uint8_t tmp;
  //ADC-Init
  ADCSRA = (1<<ADEN) | AUTO_CLOCK_DIV;	//Vorteiler=8
  lcd_init();
	

  wdt_disable();
  if(MCU_STATUS_REG & (1<<WDRF)) { 
     // berprfen auf Watchdog-Reset 
     // Das tritt ein, wenn der Watchdog 2s nicht zurckgesetzt wurde
     // Kann vorkommen, wenn sich das Programm in einer Endlosschleife "verheddert" hat.
     lcd_fix_string(TestTimedOut);	//Timeout-Meldung
     wait3s();				//warte 3 Sekunden
     ON_PORT = 0;			//Abschalten!
     return 0;
  }
  LCDLoadCustomChar(LCD_CHAR_DIODE1);	//Custom-Zeichen
  lcd_fix_customchar(DiodeIcon1);	//lade Zeichen  >|
  LCDLoadCustomChar(LCD_CHAR_DIODE2);	//Custom-Zeichen 
  lcd_fix_customchar(DiodeIcon2);	//lade Zeichen  |<
#ifdef C_MESS
  LCDLoadCustomChar(LCD_CHAR_CAP);	//Custom-Zeichen 
  lcd_fix_customchar(CapIcon);		//lade Zeichen  ||
#endif
#ifdef R_MESS
  LCDLoadCustomChar(LCD_CHAR_RESIS1);	//Custom-Zeichen 
  lcd_fix_customchar(ResIcon1);		//lade Zeichen  [
  LCDLoadCustomChar(LCD_CHAR_RESIS2);	//Custom-Zeichen 
  lcd_fix_customchar(ResIcon2);		//lade Zeichen  ]
#endif
  
#ifdef LCD_CYRILLIC
  //bei kyrillischen LCD-Zeichensatz Omega- und -Zeichen laden
  LCDLoadCustomChar(LCD_CHAR_OMEGA);	//Custom-Zeichen
  //Omega-Zeichen in LCD laden
  lcd_fix_customchar(CyrillicOmegaIcon);
  LCDLoadCustomChar(LCD_CHAR_U);	//Custom-Zeichen
  //-Zeichen in LCD laden
  lcd_fix_customchar(CyrillicMuIcon);
#endif
  Line1();	//1. Zeile

//*****************************************************************
//Einsprungspunkt, wenn Start-Taste im Betrieb erneut gedrckt wird
  start:
  PartFound = PART_NONE;
#ifdef R_MESS
  NumOfR = 0;
#endif
  NumOfDiodes = 0;
  PartReady = 0;
  PartMode = 0;
  lcd_clear();
  ADC_DDR = TXD_MSK;		//Software-UART aktivieren
#ifdef C_MESS
  ca = 0;
  cb = 0;
#endif
#ifdef PULLUP_DISABLE
  MCUCR = (1<<PUD);		// disable Pull-Up Resistors
#endif
  uart_newline();
  //Versorgungsspannung messen
  ReadADC(5);	//Dummy-Readout
  trans.hfe[0] = ReadADC(5); 	//mit 5V Referenz
    				// 5001 = 2,56V, durch Teiler 10,317V
  lcd_fix_string(Bat);		//Anzeige: "Bat. "
#ifdef BAT_OUT
  utoa(trans.hfe[0]/((100*33)/133), outval, 10);	// nur 2 Stellen ausgeben
  lcd_data(outval[0]);
  lcd_data('.');
  lcd_data(outval[1]);
  lcd_data('V');
  lcd_data(' ');
#endif
#ifdef WDT_enabled
  wdt_enable(WDTO_2S);		//Watchdog an
#endif
  if (trans.hfe[0] <  1811) {	//((7300*33)/133)
     //Vcc < 7,3V; Warnung anzeigen
     if(trans.hfe[0] < 1563) {	//((6300*33)/133)
        //Vcc <6,3V; zuverlssiger Betrieb nicht mehr mglich
        lcd_fix_string(BatEmpty);	//Batterie leer!
        wait2s();
        PORTD = 0;			//abschalten
        return 0;
     }
     lcd_fix_string(BatWeak);		//Batterie schwach
  } else { // Batterie-Spannung ausreichend
     lcd_fix_string(OK); 		// "OK"
  }
  wait1s();			// warten, damit man Batteriespannung ablesen kann
  //Test beginnen
#ifdef WITH_AUTO_REF
  RefVoltage();			//kompensiere Referenz-Spannung
#endif
#ifndef DebugOut
  Line2();
  lcd_fix_string(TestRunning);		//String: Test luft
#endif
#ifdef C_MESS
  EntladePins();		// evtl. vorhandene Restladung entladen
 #ifdef WITH_SELFTEST
  AutoCheck();			//prfe, ob Selbst-Prfung durchgefhrt werden soll
 #endif
#endif
     
  //Alle 6 Kombinationsmglichkeiten fr die 3 Pins prfen
  CheckPins(TP1, TP2, TP3);
  CheckPins(TP1, TP3, TP2);
  CheckPins(TP2, TP1, TP3);
  CheckPins(TP2, TP3, TP1);
  CheckPins(TP3, TP2, TP1);
  CheckPins(TP3, TP1, TP2);
  
#ifdef C_MESS
 #if (CapTestMode & 0x30) > 0
  //Separate Messung zum Test auf Kondensator
  if(((PartFound == PART_NONE) || (PartFound == PART_RESISTOR) || (PartFound == PART_DIODE)) ) {
     //Kondensator entladen; sonst ist evtl. keine Messung mglich
     EntladePins();		// evtl. vorhandene Restladung entladen
  #if (CapTestMode & 0x30) == 0x10
     // nur mit 2 Pins Kapazitten messen
     ReadCapacity((CapTestMode & 0x03),  (CapTestMode & 0x0c)>>2 );
  #else
     //Kapazitt in allen 6 (3) Pin-Kombinationen messen
     ReadCapacity(TP3, TP1);
     ReadCapacity(TP3, TP2);
//     ReadCapacity(TP2, TP3);
     ReadCapacity(TP2, TP1);
//     ReadCapacity(TP1, TP3);
//     ReadCapacity(TP1, TP2);
  #endif
  }
 #endif
#endif
  //Fertig, jetzt folgt die Auswertung
  lcd_clear();
  Line1();
  if(PartFound == PART_DIODE) {
     if(NumOfDiodes == 1) {		//Standard-Diode
        lcd_fix_string(Diode);		//"Diode: "
        lcd_ziff1(diodes[0].Anode);
        lcd_fix_string(AnKat);		//"->|-"
        lcd_ziff1(diodes[0].Cathode);
        UfAusgabe(0x70);
#ifdef C_MESS
        lcd_fix_string(GateCap);		//" C="
        ReadCapacity(diodes[0].Cathode,diodes[0].Anode);	// Kapazitt in Sperr-Richtung
        lcd_show_format_cap();
#endif
        goto end;
       }
     else if(NumOfDiodes == 2) { //Doppeldiode
        lcd_data('2');
        lcd_fix_string(Dioden);		//"Dioden "
        if(diodes[0].Anode == diodes[1].Anode) { //Common Anode
           lcd_ziff1(diodes[0].Cathode);
           lcd_fix_string(KatAn);	//"-|<-"
           lcd_ziff1(diodes[0].Anode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[1].Cathode);
           UfAusgabe(0x01);
           goto end;
        } else if(diodes[0].Cathode == diodes[1].Cathode) { //Common Cathode
           lcd_ziff1(diodes[0].Anode);
           lcd_fix_string(AnKat);	//"->|-"
	   lcd_ziff1(diodes[0].Cathode);
           lcd_fix_string(KatAn);	//"-|<-"
           lcd_ziff1(diodes[1].Anode);
           UfAusgabe(0x01);
           goto end;
          }
        else if ((diodes[0].Cathode == diodes[1].Anode) && (diodes[1].Cathode == diodes[0].Anode)) {
           //Antiparallel
           lcd_ziff1(diodes[0].Anode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[0].Cathode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[1].Cathode);
           UfAusgabe(0x01);
//           lcd_fix_string(Antiparallel);	//"Antiparallel"
           goto end;
        }
     } else if(NumOfDiodes == 3) {
        //Serienschaltung aus 2 Dioden; wird als 3 Dioden erkannt
        trans.b = 3;
        trans.c = 3;
        /* berprfen auf eine fr eine Serienschaltung von 2 Dioden mgliche Konstellation
          Es stimmt nur einmal die Pin-Nr. einer Kathode mit einer Anode berein.
          Das kommmt daher, dass die Dioden als 2 Einzeldioden und ZUSTZLICH als eine "groe" Diode erkannt werden.
        */
        if(diodes[0].Cathode == diodes[1].Anode)
          {
           trans.b = 0;
           trans.c = 1;
          }
        if(diodes[0].Anode == diodes[1].Cathode)
          {
           trans.b = 1;
           trans.c = 0;
          }
        if(diodes[0].Cathode == diodes[2].Anode)
          {
           trans.b = 0;
           trans.c = 2;
          }
        if(diodes[0].Anode == diodes[2].Cathode)
          {
           trans.b = 2;
           trans.c = 0;
          }
        if(diodes[1].Cathode == diodes[2].Anode)
          {
           trans.b = 1;
           trans.c = 2;
          }
        if(diodes[1].Anode == diodes[2].Cathode)
          {
           trans.b = 2;
           trans.c = 1;
          }
#if DebugOut == 4
	Line3();
        lcd_ziff1(diodes[0].Anode);
        lcd_data(':');
        lcd_ziff1(diodes[0].Cathode);
        lcd_data(' ');
        lcd_string(utoa(diodes[0].Voltage, outval, 10));
        lcd_data(' ');
        lcd_ziff1(diodes[1].Anode);
        lcd_data(':');
        lcd_ziff1(diodes[1].Cathode);
        lcd_data(' ');
        lcd_string(utoa(diodes[1].Voltage, outval, 10));
	Line4();
        lcd_ziff1(diodes[2].Anode);
        lcd_data(':');
        lcd_ziff1(diodes[2].Cathode);
        lcd_data(' ');
        lcd_string(utoa(diodes[2].Voltage, outval, 10));
        Line1();
#endif
        if((trans.b<3) && (trans.c<3)) {
           lcd_data('3');
           lcd_fix_string(Dioden);	//"Dioden "
           lcd_ziff1(diodes[trans.b].Anode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[trans.b].Cathode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[trans.c].Cathode);
           UfAusgabe( (trans.b<<4)|trans.c);
           goto end;
        }
     }
     // end (PartFound == PART_DIODE)
  } else if (PartFound == PART_TRANSISTOR) {
    if(PartReady == 0) {	//Wenn 2. Prfung nie gemacht, z.B. bei Transistor mit Schutzdiode
       trans.hfe[1] = trans.hfe[0];
    }
    if((trans.hfe[0]>trans.hfe[1])) {
       //Wenn der Verstrkungsfaktor beim ersten Test hher war: C und E vertauschen!
       trans.hfe[1] = trans.hfe[0];
       tmp = trans.c;
       trans.c = trans.e;
       trans.e = tmp;
    }

    if(PartMode == PART_MODE_NPN) {
       lcd_fix_string(NPN);		//"NPN "
    } else {
       lcd_fix_string(PNP);		//"PNP "
    }
    if(NumOfDiodes > 2) {	//Transistor mit Schutzdiode
       if(PartMode == PART_MODE_NPN) {
          lcd_fix_string(AnKat);	//"->|-"
       } else {
          lcd_fix_string(KatAn);	//"-|<-"
       }
    }
    lcd_fix_string(ebcstr);		//" EBC="
    lcd_ziff1(trans.e);
    lcd_ziff1(trans.b);
    lcd_ziff1(trans.c);
    Line2(); //2. Zeile
    lcd_fix_string(hfestr);		//"hFE="
    lcd_string(utoa(trans.hfe[1], outval, 10));
    lcd_data(' ');

    for(trans.c=0;trans.c<NumOfDiodes;trans.c++) {
       if(((diodes[trans.c].Cathode == trans.e) && (diodes[trans.c].Anode == trans.b) && (PartMode == PART_MODE_NPN)) ||
          ((diodes[trans.c].Anode == trans.e) && (diodes[trans.c].Cathode == trans.b) && (PartMode == PART_MODE_PNP))) {
          lcd_fix_string(Uf);		//"Uf="
          lcd_string(utoa(diodes[trans.c].Voltage, outval, 10));
          lcd_data('m');
          lcd_data('V');
          goto end;
       }
    }
    goto end;
    // end (PartFound == PART_TRANSISTOR)
  } else if (PartFound == PART_FET) {	//JFET oder MOSFET
    if(PartMode&1) {
       //N-Kanal
       lcd_data('N');
    } else {
       lcd_data('P');			//P-Kanal
    }
    lcd_data('-');
    if((PartMode==PART_MODE_N_D_MOS) || (PartMode==PART_MODE_P_D_MOS)) {
       lcd_data('D');			// N-D
       lcd_fix_string(mosfet);		//"-MOS"
    } else {
       if((PartMode==PART_MODE_N_JFET) || (PartMode==PART_MODE_P_JFET)) {
          lcd_fix_string(jfet);		//"-JFET"
       } else {
          lcd_data('E');		// -E
          lcd_fix_string(mosfet);	//"-MOS"
       }
    }
 #ifdef C_MESS	//Gatekapazitt
    if(PartMode < 3) {  //Anreicherungs-MOSFET
       lcd_fix_string(GateCap);		//" C="
       ReadCapacity(trans.b,trans.e);	//Messung Kapazitt
       lcd_show_format_cap();
    }
 #endif
    Line2(); //2. Zeile
    lcd_fix_string(gds);		//"GDS="
    lcd_ziff1(trans.b);
    lcd_ziff1(trans.c);
    lcd_ziff1(trans.e);
    if((NumOfDiodes > 0) && (PartMode < 3)) {
       //MOSFET mit Schutzdiode; gibt es nur bei Anreicherungs-FETs
       lcd_data(LCD_CHAR_DIODE1);	//Diode anzeigen >|
    } else {
       lcd_data(' ');	//Leerzeichen
    }
    if(PartMode < 3) {	//Anreicherungs-MOSFET
       lcd_fix_string(vt);
       lcd_string(utoa(gthvoltage, outval, 10));	//Gate-Schwellspannung, wurde zuvor ermittelt
       lcd_data('m');
    }
    goto end;
    // end (PartFound == PART_FET)
  } else if (PartFound == PART_THYRISTOR) {
    lcd_fix_string(Thyristor);		//"Thyristor"
    goto gakAusgabe;
  } else if (PartFound == PART_TRIAC) {
    lcd_fix_string(Triac);		//"Triac"
    goto gakAusgabe;
  }
 #ifdef R_MESS	//Widerstandsmessung nur mit Mega8 verfgbar
  else if(PartFound == PART_RESISTOR) {
//    lcd_fix_string(Resistor); 		//"Widerst: "
    if (NumOfR == 1) { // Einzelwiderstand
       lcd_ziff1(resis[ii].rb);  	//Pin-Angabe
       lcd_fix_string(Resis);
       lcd_ziff1(resis[ii].ra);		//Pin-Angaben
    } else { // R-Max suchen
       ii = 0;
       if ((resis[1].rx * resis[1].kfl) > (resis[0].rx * resis[0].kfl))
          ii = 1;
       if (NumOfR == 2) {
          ii = 2;
       } else {
          if ((resis[2].rx * resis[2].kfl) > (resis[ii].rx * resis[ii].kfl)) {
             ii = 2;
          }
       }

       if (ii == 0) {
          lcd_data('1');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('3');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('2');
       }
       if (ii == 1) {
          lcd_data('1');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('2');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('3');
       }
       if (ii == 2) {
          lcd_data('2');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('1');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('3');
       }
    }
    Line2(); //2. Zeile
    if (NumOfR == 1) {
       RvalOut(0);
    } else {
       if (ii == 0) {
          RvalOut(1);
          RvalOut(2);
       }
       if (ii == 1) {
          RvalOut(0);
          RvalOut(2);
       }
       if (ii == 2) {
          RvalOut(0);
          RvalOut(1);
       }
    }
    goto end;

  } // end (PartFound == PART_RESISTOR)
#endif

#ifdef C_MESS
//Kapazittsmessung auch nur auf Mega8 verfgbar
  else if(PartFound == PART_CAPACITOR) {
//     lcd_fix_string(Capacitor);
     lcd_ziff1(ca);			//Pin-Angaben
     lcd_fix_string(CapZeich);
     lcd_ziff1(cb);
     Line2(); 				//2. Zeile
     lcd_show_format_cap();
     goto end;
  }
#endif
  if(NumOfDiodes == 0) { //Keine Dioden gefunden
     lcd_fix_string(TestFailed1); 	//"Kein,unbek. oder"
     Line2(); //2. Zeile
     lcd_fix_string(TestFailed2); 	//"defektes "
     lcd_fix_string(Bauteil);		//"Bauteil"
  } else {
     lcd_fix_string(Bauteil);		//"Bauteil"
     lcd_fix_string(Unknown); 		//" unbek."
     Line2(); //2. Zeile
     lcd_fix_string(OrBroken); 		//"oder defekt "
     lcd_data(NumOfDiodes + '0');
     lcd_fix_string(AnKat);		//"->|-"
  }
  goto end;


gakAusgabe:
  Line2(); //2. Zeile
  lcd_fix_string(GAK);		//"GAK="
  lcd_ziff1(trans.b);
  lcd_ziff1(trans.c);
  lcd_ziff1(trans.e);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - 
 end:
  while(!(ON_PIN_REG & (1<<RST_PIN)));	//warten ,bis Taster losgelassen
  wait200ms();
// 10 Sekunden warten
  for(gthvoltage = 0;gthvoltage<10000;gthvoltage++) {
     if(!(ON_PIN_REG & (1<<RST_PIN))) {
        // Wenn der Taster wieder gedrckt wurde...
        // wieder zum Anfang springen und neuen Test durchfhren
        goto start;
     }
     wdt_reset();
     wait1ms();
  }
  wdt_disable();			//Watchdog aus
  ON_PORT &= ~(1<<ON_PIN);		//Abschalten
  //Endlosschleife
  while(1) {
     if(!(ON_PIN_REG & (1<<RST_PIN))) {
        // wird nur erreicht,
        // wenn die automatische Abschaltung nicht eingebaut wurde 
        goto start;
     }
  }
  return 0;
}   // end main

//******************************************************************
// Ausgabe der Flussspannung fr 1-2 Dioden in Zeile 2
// bcdnum = Nummern der beiden Dioden:
// oberer 4 Bit  Index der ersten Diode
// untere 4 Bit  Index der zweiten Diode (Struktur diodes[nn])
// wenn Index >= 3 erfolgt keine Ausgabe
void UfAusgabe(uint8_t bcdnum) {
   uint8_t nn;

   Line2(); 				//2. Zeile
   lcd_fix_string(Uf);			//"Uf="
   nn = (bcdnum >> 4);
   if (nn < 3) {
      lcd_string(utoa(diodes[nn].Voltage, outval, 10));
      lcd_data('m');
      lcd_data('V');
      lcd_data(' ');
   }
   nn = (bcdnum & 0x0f);
   if (nn < 3) {
      lcd_string(utoa(diodes[nn].Voltage, outval, 10));
      lcd_data('m');
      lcd_data('V');
   }
}

#ifdef R_MESS
void RvalOut(uint8_t ii) {	
   // Widerstandswert ausgeben
   uint8_t strlength;
   char prefix;

   ultoa(resis[ii].rx,outval,10);
   strlength = strlen(outval);	//Ntig, um Komma anzuzeigen

   if (resis[ii].kfl == 100) {  //470k-Widerstand?
      if (strlength < 5) {
         outval[strlength] = outval[strlength-1];
         outval[strlength-1] = '.';
         outval[strlength+1] = (char)0;
         prefix = 'k';
      }
      if (strlength == 5) { 
         outval[4] = outval[3];
         outval[3] = outval[2];
         outval[2] = outval[1];
         outval[1] = '.';
         outval[5] = (char)0;
         prefix = 'M';
      }
      if (strlength == 6) {  
         outval[4] = outval[3];
         outval[3] = outval[2];
         outval[2] = '.';
         outval[5] = (char)0;
         prefix = 'M';
      }
   } else {
      prefix = (char)0;
//      if (strlength == 4) {
//         outval[4] = outval[3];
//         outval[3] = outval[2];
//         outval[2] = outval[1];
//         outval[1] = '.';
//         outval[5] = (char)0;
//         prefix = 'k';
//      }
      if (strlength == 5) {
         outval[4] = outval[3];
         outval[3] = outval[2];
         outval[2] = '.';
         outval[5] = (char)0;
         prefix = 'k';
      }
   }
   lcd_string(outval);			//modifizierten String ausgeben
   if (prefix != (char)0) {
      lcd_data(prefix);
   }
   lcd_data(LCD_CHAR_OMEGA);	//Omega fr Ohm 
   lcd_data(' ');
 }
#endif

//******************************************************************
#include "CheckPins.c"


void ChargePin10ms(uint8_t PinToCharge, uint8_t ChargeDirection) {
   //Anschluss eines Bauelementes kurz(10ms) auf ein bestimmtes Potenzial legen
   //Diese Funktion ist zum Entladen von MOSFET-Gates vorgesehen, um Schutzdioden u.. in MOSFETs
   //erkennen zu knnen
   //Parameter:
   //PinToCharge: zu entladender Pin als Bits fr den R-Port
   //ChargeDirection: 0 = gegen Masse (N-Kanal-FET), 1= gegen Plus(P-Kanal-FET)

   if(ChargeDirection&1) {
      R_PORT |= PinToCharge;	//R_PORT auf 1 (VCC) 
   } else {
      R_PORT &= ~PinToCharge; // oder 0 (GND)
   }
   R_DDR |= PinToCharge;			//Pin auf Ausgang, ber R auf Masse oder VCC
   wait10ms();
   // hochohmig schalten, Input
   R_DDR &= ~PinToCharge;	// R fr Pin wieder auf Eingang
   R_PORT &= ~PinToCharge;	// kein Pullup
}

// erst vorhandene Restladung entladen
void EntladePins() {
  uint8_t adc_gnd;		// Maske der ADC-Ausgnge, die direkt auf GND gelegt werden
  unsigned int adcmv[3];	// Spannungswerte der 3 Pins in mV
  unsigned int clr_cnt;		// Clear Counter
  uint8_t lop_cnt;
// Entladezeit in ms / Wartezeit (10000/20)
#define MAX_ENTLADE_ZEIT  (10000/20)

  for(lop_cnt=0;lop_cnt<10;lop_cnt++) {
     adc_gnd = TXD_MSK;		// alle ADC auf Input
     ADC_DDR = adc_gnd;
     ADC_PORT = TXD_VAL;		// ADC-Ausgnge auf 0
     R_PORT = 0;			// R-Ausgnge auf 0
     R_DDR = (2<<(PC2*2)) | (2<<(PC1*2)) | (2<<(PC0*2)); // alle Pins ber R_H auf Masse
     adcmv[0] = W5msReadADC(PC0);	// welche Spannung hat Pin 1?
     adcmv[1] = ReadADC(PC1);	// welche Spannung hat Pin 2?
     adcmv[2] = ReadADC(PC2);	// welche Spannung hat Pin 3?
     if ((adcmv[0] < 3) & (adcmv[1] < 3) & (adcmv[2] < 3)) {
        ADC_DDR = TXD_MSK;		// alle direkten ADC-Ausgnge hochohmig
        R_DDR = 0;			// alle R_L Ports (und R_H) hochohmig
        return;
     }
     // alle Pins mit weniger als 1V direkt ber ADC-Port auf Masse (GND) schalten
     if (adcmv[0] < 1000) {
        adc_gnd |= (1<<PC0);	//Pin 1 ber Ausgang des Analog-Pins auf GND
     }
     if (adcmv[1] < 1000) {
        adc_gnd |= (1<<PC1);	//Pin 2 ber Ausgang des Analog-Pins auf GND
     }
     if (adcmv[2] < 1000) {
        adc_gnd |= (1<<PC2);	//Pin 3 ber Ausgang des Analog-Pins auf GND
     }
     ADC_DDR = adc_gnd;		// jetzt alle gewhlten ADC-Ports auf GND schalten

     // zustzlich die verbleibenden Ports ber R_L auf Masse.
     // Da es bei den direkt auf Masse geschalteten Pins auch nicht schadet, werden
     // einfach alle R_L Widerstnde geschaltet
     R_DDR = (1<<(PC2*2)) | (1<<(PC1*2)) | (1<<(PC0*2));	// Pins ber R_L Widerstnde auf GND
     for(clr_cnt=0;clr_cnt<MAX_ENTLADE_ZEIT;clr_cnt++) {
        wdt_reset();
        adcmv[0] = W20msReadADC(PC0);	// welche Spannung hat Pin 1?
        adcmv[1] = ReadADC(PC1);	// welche Spannung hat Pin 2?
        adcmv[2] = ReadADC(PC2);	// welche Spannung hat Pin 3?
        if (adcmv[0] < 1300) {
           ADC_DDR |= (1<<PC0);	// unter 1.3V ber ADC-Port auf GND
        }
        if (adcmv[1] < 1300) {
           ADC_DDR |= (1<<PC1);	// unter 1.3V ber ADC-Port auf GND
        }
        if (adcmv[2] < 1300) {
           ADC_DDR |= (1<<PC2);	// unter 1.3V ber ADC-Port auf GND
        }
        if ((adcmv[0] < 5) && (adcmv[1] < 5) && (adcmv[2] < 5)) {
           break;
        }
     }
     if (clr_cnt == MAX_ENTLADE_ZEIT) {
        // Kondensator hat Restladung, spter warnen!
//        Line3();
//        lcd_data('E');
//        lcd_string(utoa(adcmv[0], outval, 10));
//        lcd_data(' ');
//        lcd_string(utoa(adcmv[1], outval, 10));
//        lcd_data(' ');
//        lcd_string(utoa(adcmv[2], outval, 10));
     }
     R_PORT = 0;		// alle R_Ausgnge auf 0
     for(adcmv[0]=0;adcmv[0]<clr_cnt;adcmv[0]++) {
        // Zur Sicherheit nochmal 1/20 der Zeit im Kurzschluss-Mode
        wait1ms();
     }
  } // end while
 }




#ifdef C_MESS	//Kapazittsmessung nur auf Mega8 verfgbar
#include "ReadCapacity.c"


void lcd_show_format_cap(void) {
  uint8_t strlength;
  while(cval > 399999) {	//ab 400nF oder 400F
//   cval /= 1000;
     cval = (cval + 500) / 1000;	// mit Rundung
     cpre++;				// F-prefix um 3 Zehnerpotenzen erhhen
  }
  ultoa(cval, outval, 10);
  strlength = strlen(outval);
  if (strlength == 6) { // letzten 2 Stellen weglassen
     outval[4] = outval[3];		// Platz schaffen fr '.'
     outval[3] = '.';
     cpre++;				// F-prefix um 3 Zehnerpotenzen erhhen
  }
  if (strlength == 5) {
     outval[4] = outval[3];		// Platz schaffen fr '.'
     outval[3] = outval[2];
     outval[2] = '.';
     cpre++;				// F-prefix um 3 Zehnerpotenzen erhhen
  }
  outval[5] = 0;

  lcd_string(outval);
  tmpval2 = pgm_read_byte(&C_Prefix_tab[cpre]);
  lcd_data(tmpval2);		// p oder n oder  oder m
  lcd_data('F');
 }
unsigned int getRLmultip(unsigned int cvolt) {
// Widerstand 680 Ohm         200  250  300  350  400  450  500  550  600  650  700  750  800  850  900  950 1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
//uint16_t RLtab[] MEM_TEXT = {3402,2708,2245,1914,1666,1473,1318,1192,1086, 997, 921, 855, 797, 745, 700, 659, 622, 589, 559, 531, 506, 483, 461, 441, 423};
#define Tab_Abstand 50		// Tabellenabstand ist 50mV
#define Tab_RL_Beginn 200	// Tabellenbeginn ist 200mV
  unsigned int uvolt;
  unsigned int y1, y2;
  uint8_t tabind;
  uint8_t tabres;
  if (cvolt >= Tab_RL_Beginn) {
     uvolt = cvolt - Tab_RL_Beginn;
  } else {
     uvolt = 0;		// begrenze auf Tabellenbeginn
  }
  tabind = uvolt / Tab_Abstand;
  tabres = uvolt % Tab_Abstand;
  tabres = Tab_Abstand - tabres;
  if (tabind > 23) {
     tabind = 23;
  }
  #ifdef USE_EEPROM
  y1 = eeprom_read_word(&RLtab[tabind]);
  y2 = eeprom_read_word(&RLtab[tabind+1]);
  #else
  y1 = pgm_read_word(&RLtab[tabind]);
  y2 = pgm_read_word(&RLtab[tabind+1]);
  #endif
  return ( ((y1 - y2) * tabres + (Tab_Abstand/2)) / Tab_Abstand + y2);
}


 #ifdef WITH_AUTO_REF
void RefVoltage(void) {
// zur Erinnerung:
// Widerstand 680 Ohm        1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
//uint16_t RLTAB[] MEM_TEXT = { 702, 660, 622, 588, 558, 529, 504, 480, 458};
//Widerstand 470000 Ohm      1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
//uint16_t RHTAB[] MEM_TEXT = { 954, 903, 856, 814, 775, 740, 707, 676, 648};
#define Tab_Abstand 50		// Tabellenabstand ist 50mV
#define Tab_Beginn 1000		// Tabelle beginnt bei 1000mV

  unsigned int ref_mv;
  unsigned int y1, y2;
  uint8_t tabind;
  uint8_t tabres;

  ref_mv = ReadADC(0x0e);	// Referenz-Spannung lesen
  ref_mv = W5msReadADC(0x0e);	// Referenz-Spannung lesen
  if (ref_mv >= Tab_Beginn) {
     ref_mv -= Tab_Beginn;
  } else  {
     ref_mv = 0;		// Begrenze auf Tabellenbeginn
  }
  tabind = ref_mv / Tab_Abstand;
  tabres = ref_mv % Tab_Abstand;
  tabres = Tab_Abstand-tabres;
  if (tabind > 7) {
     tabind = 7;
  }
  // Interpoliere die Tabelle der Faktoren
  #ifdef USE_EEPROM
  y1 = eeprom_read_word(&RHtab[tabind]);
  y2 = eeprom_read_word(&RHtab[tabind+1]);
  #else
  y1 = pgm_read_word(&RHtab[tabind]);
  y2 = pgm_read_word(&RHtab[tabind+1]);
  #endif
  // RHmultip ist der interpolierte Faktor fr 470k
  RHmultip = ((y1 - y2) * tabres + (Tab_Abstand/2)) / Tab_Abstand + y2;
  #ifdef USE_EEPROM
  y1 = eeprom_read_word(&RLtab[tabind]);
  y2 = eeprom_read_word(&RLtab[tabind+1]);
  #else
  y1 = pgm_read_word(&RLtab[tabind]);
  y2 = pgm_read_word(&RLtab[tabind+1]);
  #endif
  // RLmultip ist der interpolierte Faktor fr 680
  RLmultip = ((y1 - y2) * tabres + (Tab_Abstand/2)) / Tab_Abstand + y2;
 }
 #endif
#endif

#ifdef LCD_CLEAR
void lcd_clear_line(void) {
 // Schreibt 20 Leerzeichen auf Display, Cursor muss am Anfang der Zeile stehen
 unsigned char ll;
 for (ll=0;ll<20;ll++) {
    lcd_data(' ');
 }
}
#endif

#ifdef WITH_SELFTEST
void AutoCheck(void) {
  uint8_t tt;		// laufende Testnummer
  uint8_t ww;		// Zhler fr Testwiederholung
  unsigned int  adcmv[3];
  unsigned int ref_mv;
  ADC_PORT = TXD_VAL;
  ADC_DDR = TXD_MSK;
  R_PORT = (1<<(TP1*2));		//Pin 1 ber RL auf +
  R_DDR = (1<<(TP1*2)) | (1<<(TP2*2));	//Pin 2 ber RL auf -
  adcmv[0] = W5msReadADC(PC0);		// Spannung an Pin 1
  adcmv[1] = ReadADC(PC1);		// Spannung an Pin 2
  adcmv[2] = ReadADC(PC2);		// Spannung an Pin 3
  if (abs(adcmv[1] - adcmv[0]) > 10) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs(adcmv[2] - adcmv[0]) > 10) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs(adcmv[0] - (U_VCC/2)) > 20) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  R_DDR = (1<<(TP1*2)) | (1<<(TP3*2));	//Pin 3 ber RL auf - (Pin 1 ber RL auf +)
  adcmv[0] = W5msReadADC(PC0);		// Spannung an Pin 1
  adcmv[1] = ReadADC(PC1);		// Spannung an Pin 2
  adcmv[2] = ReadADC(PC2);		// Spannung an Pin 3
  if (abs(adcmv[1] - adcmv[0]) > 10) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs(adcmv[2] - adcmv[0]) > 10) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs(adcmv[0] - (U_VCC/2)) > 20) {
     return;				//Abweichung zu gross, kein Selbsttest
  }

  lcd_clear();
  Line1();
  lcd_fix_string(SELFTEST);		// "Selftest mode.."
  wait1s();
 
  for(tt=1;tt<10;tt++) {			// Scheife ber Tests
     for(ww=0;ww<8;ww++) {		// Testwiederholung
        Line1();			//Cursor auf Spalte1, Zeile1
        lcd_clear_line();
        Line1();			//Cursor auf Spalte1, Zeile1
        lcd_data('T');
        lcd_data('0'+tt);
        lcd_data(' ');
        if (tt == 1) {   // Ausgabe der Referenzspannung und Faktoren fr C-Messung
           ref_mv = ReadADC(0x0e);	// Referenz-Spannung lesen
           ref_mv = W5msReadADC(0x0e);	// Referenz-Spannung lesen
           lcd_fix_string(URefT);	//"URef="
           lcd_string(utoa(ref_mv, outval, 10));
           lcd_fix_string(mVT);		//"mV "
           Line2();
           lcd_clear_line();
           Line2();
           lcd_fix_string(RHfakt);	//"RHf="
           lcd_string(utoa(RHmultip, outval, 10));
           lcd_fix_string(RLfakt);	//" RLf="
#ifdef C_MESS
           lcd_string(utoa(getRLmultip(ref_mv), outval, 10));
#else
           lcd_string(utoa(RLmultip, outval, 10));
#endif
        }
        if (tt == 2) { // Messung gleiche Widerstnde RL
           R_PORT = 1<<(TP1*2);		//RL1 auf VCC
           R_DDR = (1<<(TP1*2)) | (1<<(TP2*2));	//RL2 auf -
           adcmv[0] = W20msReadADC(TP1);
           R_DDR = (1<<(TP1*2)) | (1<<(TP3*2));	//RL3 auf -
           adcmv[1] = W20msReadADC(TP1);
           R_PORT = 1<<(TP2*2);		//RL2 auf VCC
           R_DDR = (1<<(TP2*2)) | (1<<(TP3*2));	//RL3 auf -
           adcmv[2] = W20msReadADC(TP1);
           lcd_fix_string(RLRL);	// "RLRL"
        }
        if (tt == 3) { // Messung gleiche Widerstnde RH
           R_PORT = 2<<(TP1*2);		//RH1 auf VCC
           R_DDR = (2<<(TP1*2)) | (2<<(TP2*2));	//RH2 auf -
           adcmv[0] = W20msReadADC(TP1);
           R_DDR = (2<<(TP1*2)) | (2<<(TP3*2));	//RH3 auf -
           adcmv[1] = W20msReadADC(TP1);
           R_PORT = 2<<(TP2*2);		//RL2 auf VCC
           R_DDR = (2<<(TP2*2)) | (2<<(TP3*2));	//RH3 auf -
           adcmv[2] = W20msReadADC(TP1);
           lcd_fix_string(RHRH);	// "RHRH"
        }
        if (tt == 4) { // Text trennen
           lcd_fix_string(RELPROBE);	// "Release Probes"
        }
        if (tt == 5) { // Messung, wie gut R_H die Spannung auf Ground und VCC zieht
           R_PORT = 0;
           R_DDR = 2<<(TP1*2);		//Pin 1 ber R_H auf Masse
           adcmv[0] = W20msReadADC(TP1);

           R_DDR = 2<<(TP2*2);		//Pin 2 ber R_H auf Masse
           adcmv[1] = W20msReadADC(TP2);

           R_DDR = 2<<(TP3*2);		//Pin 3 ber R_H auf Masse
           adcmv[2] = W20msReadADC(TP3);
           lcd_fix_string(RH1L);	// "RH_Lo="
           Line2();
        }
        if (tt == 6) { // Messung, wie gut R_H die Spannung auf VCC zieht
           R_DDR = 2<<(TP1*2);		//Pin 1 ber R_H auf VCC
           R_PORT = 2<<(TP1*2);
           adcmv[0] = W20msReadADC(TP1);
           R_DDR = 2<<(TP2*2);		//Pin 2 ber R_H auf VCC
           R_PORT = 2<<(TP2*2);
           adcmv[1] = W20msReadADC(TP2);
           R_DDR = 2<<(TP3*2);		//Pin 3 ber R_H auf VCC
           R_PORT = 2<<(TP3*2);
           adcmv[2] = W20msReadADC(TP3);
           lcd_fix_string(RH1H);	// "RH_Hi="
        }
        if (tt == 7) { // Messung des Innenwiderstandes der Ports
           ADC_DDR = 1<<TP1 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_PORT = 1<<(TP1*2); // Alle R_L-PORTS auf VCC
           R_DDR = 1<<(TP1*2);		//Low-Pin auf Ausgang und ber R_L auf VCC
           adcmv[0] = W5msReadADC(TP1);

           ADC_DDR = 1<<TP2 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_PORT =  1<<(TP2*2); // Alle R_L-PORTS auf VCC
           R_DDR = 1<<(TP2*2);		//High-Pin auf Ausgang und ber R_L auf VCC
           adcmv[1] = W5msReadADC(TP2);

           ADC_DDR = 1<<TP3 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_PORT =  1<<(TP3*2); // Alle R_L-PORTS auf VCC
           R_DDR = 1<<(TP3*2);		//High-Pin auf Ausgang und ber R_L auf VCC
           adcmv[2] = W5msReadADC(TP3);
           lcd_fix_string(RILO);	// "RiLo="
        }
        if (tt == 8) { // Messung des Innenwiderstandes der Ports
           R_PORT = 0;			// alle R-Ports auf Masse
           ADC_PORT = 1<<TP1 | TXD_VAL;	// alle ADC-Ports auf VCC
           ADC_DDR = 1<<TP1 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_DDR = 1<<(TP1*2);			//Low-Pin auf Ausgang und ber R_L auf Masse
           adcmv[0] = U_VCC - W5msReadADC(TP1);
      
           ADC_PORT = 1<<TP2 | TXD_VAL;	// alle ADC-Ports auf VCC
           ADC_DDR = 1<<TP2 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_DDR = 1<<(TP2*2);		//High-Pin auf Ausgang und ber R_L auf Masse
           adcmv[1] = U_VCC - W5msReadADC(TP2);

           ADC_PORT = 1<<TP3 | TXD_VAL;	// alle ADC-Ports auf VCC
           ADC_DDR = 1<<TP3 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_DDR = 1<<(TP3*2);		//High-Pin auf Ausgang und ber R_L auf Masse
           adcmv[2] = U_VCC - W5msReadADC(TP3);

           lcd_fix_string(RIHI);	// "RiHi="
        }
        if (tt == 9) { 			//Frequenz-Generator 50Hz
           ADC_PORT = TXD_VAL;
           ADC_DDR = 1<<TP1 | TXD_MSK;	// Pin 1 auf GND
           R_DDR = 1<<TP3;
           for (ii=0;ii<500;ii++) {
              R_PORT = 0;
              wait10ms();
              R_PORT = 1<<TP3;
              wait10ms();
              wdt_reset();
           }
        }
        if (tt > 1) {			// die gemessenen 3 Spannungen ausgeben
           Line2();			//Cursor auf Spalte1, Zeile2
           lcd_clear_line();
           Line2();
           lcd_string(utoa(adcmv[0], outval, 10));
           lcd_data(' ');
           lcd_string(utoa(adcmv[1], outval, 10));
           lcd_data(' ');
           lcd_string(utoa(adcmv[2], outval, 10));
        }
        ADC_DDR =  TXD_MSK;		//alle-Pins auf Input
        ADC_PORT = TXD_VAL;		// alle ADC-Ports auf GND
        R_DDR = 0;			// alle R-Ports auf Input
        R_PORT = 0;
        wait1s();
     } //end for ww
     wait1s();
  } //end for tt
  lcd_clear();
  Line1();
  lcd_fix_string(ATE);		//"Auto Test End"
  Line2();
  lcd_fix_string(VERSION);	//"Version ..."
  wait2s();			//warte 2 Sekunden
 }
#endif
